"""
DBT Function specification module.
"""
from __future__ import annotations

from digitalhub_core.entities.functions.spec import FunctionParams, FunctionSpec
from digitalhub_core.utils.exceptions import EntityError
from digitalhub_core.utils.generic_utils import decode_string, encode_string


class FunctionSpecDBT(FunctionSpec):
    """
    Specification for a Function DBT.
    """

    def __init__(
        self,
        source: str | None = None,
        sql: str | None = None,
        **kwargs,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        sql : str
            SQL query to run inside DBT.
        """
        super().__init__(source, **kwargs)
        if sql is None:
            raise EntityError("SQL query must be provided.")

        # This is to avoid re-encoding the SQL query when
        # it is already encoded.
        try:
            sql = decode_string(sql)
        except Exception:
            ...
        self.sql = encode_string(sql)


class FunctionParamsDBT(FunctionParams):
    """
    Function DBT parameters model.
    """

    sql: str = None
    """SQL query to run inside the container."""



import os
from pathlib import Path

####################
# Templates
####################

PROJECT_TEMPLATE = """
name: "{}"
version: "1.0.0"
config-version: 2
profile: "postgres"
model-paths: ["{}"]
models:
"""

MODEL_TEMPLATE_UUID = """
models:
  - name: {}
    latest_uuid: {}
    uuids:
        - v: {}
          config:
            materialized: table
"""

MODEL_TEMPLATE_VERSION = """
models:
  - name: {}
    latest_version: {}
    versions:
        - v: {}
          config:
            materialized: table
"""

PROFILE_TEMPLATE = f"""
postgres:
    outputs:
        dev:
            type: postgres
            host: {os.getenv("POSTGRES_HOST")}
            user: {os.getenv("POSTGRES_USER")}
            pass: {os.getenv("POSTGRES_PASSWORD")}
            port: {os.getenv("POSTGRES_PORT")}
            dbname: {os.getenv("POSTGRES_DATABASE")}
            schema: {os.getenv("POSTGRES_SCHEMA", "public")}
    target: dev
"""

####################
# Functions
####################


def generate_dbt_profile_yml(root_dir: Path) -> None:
    """
    Create dbt profiles.yml

    Returns
    -------
    None
    """
    profiles_path = root_dir / "profiles.yml"
    profiles_path.write_text(PROFILE_TEMPLATE)


def generate_dbt_project_yml(root_dir: Path, model_dir: Path, project: str) -> None:
    """
    Create dbt_project.yml from 'dbt'

    Parameters
    ----------
    project : str
        The project name.

    Returns
    -------
    None
    """
    project_path = root_dir / "dbt_project.yml"
    project_path.write_text(PROJECT_TEMPLATE.format(project, model_dir.name))


def generate_outputs_conf(model_dir: Path, sql: str, output: str, uuid: str) -> None:
    """
    Write sql code for the model and write schema
    and version detail for outputs versioning

    Parameters
    ----------
    sql : str
        The sql code.
    output : str
        The output table name.
    uuid : str
        The uuid of the model for outputs versioning.

    Returns
    -------
    None
    """
    sql_path = model_dir / f"{output}.sql"
    sql_path.write_text(sql)

    output_path = model_dir / f"{output}.yml"
    output_path.write_text(MODEL_TEMPLATE_VERSION.format(output, uuid, uuid))


def generate_inputs_conf(model_dir: Path, name: str, uuid: str) -> None:
    """
    Generate inputs confs dependencies for dbt project.

    Parameters
    ----------
    project : str
        The project name.
    inputs : list
        The list of inputs dataitems names.

    Returns
    -------
    None
    """
    # write schema and version detail for inputs versioning
    input_path = model_dir / f"{name}.sql"
    input_path.write_text(MODEL_TEMPLATE_UUID.format(name, uuid, uuid))

    # write also sql select for the schema
    sql_path = model_dir / f"{name}_v{uuid}.sql"
    sql_path.write_text(f'SELECT * FROM "{name}_v{uuid}"')



"""
Runtime DBT parsing utils module.
"""
from __future__ import annotations

import typing
from dataclasses import dataclass

from dbt.cli.main import dbtRunnerResult
from digitalhub_core.utils.generic_utils import encode_string
from digitalhub_core.utils.logger import LOGGER

if typing.TYPE_CHECKING:
    from dbt.contracts.results import RunResult


@dataclass
class ParsedResults:
    """
    Parsed results class.
    """

    name: str
    path: str
    raw_code: str
    compiled_code: str
    timings: dict


def parse_results(run_result: dbtRunnerResult, output: str, project: str) -> ParsedResults:
    """
    Parse dbt results.

    Parameters
    ----------
    run_result : dbtRunnerResult
        The dbt result.
    output : str
        The output table name.
    project : str
        The project name.

    Returns
    -------
    ParsedResults
        Parsed results.
    """
    result: RunResult = validate_results(run_result, output, project)
    try:
        path = get_path(result)
        raw_code = get_raw_code(result)
        compiled_code = get_compiled_code(result)
        timings = get_timings(result)
        name = result.node.name
        return ParsedResults(name, path, raw_code, compiled_code, timings)
    except Exception:
        msg = "Something got wrong during results parsing."
        LOGGER.exception(msg)
        raise RuntimeError(msg)


def validate_results(run_result: dbtRunnerResult, output: str, project: str) -> RunResult:
    """
    Parse dbt results.

    Parameters
    ----------
    run_result : dbtRunnerResult
        The dbt result.
    output : str
        The output table name.
    project : str
        The project name.

    Returns
    -------
    RunResult
        Run result.

    Raises
    ------
    RuntimeError
        If something got wrong during function execution.
    """
    try:
        # Take last result, final result of the query
        result: RunResult = run_result.result[-1]
    except IndexError:
        msg = "No results found."
        LOGGER.error(msg)
        raise RuntimeError(msg)

    if not result.status.value == "success":
        msg = f"Function execution failed: {result.status.value}."
        LOGGER.error(msg)
        raise RuntimeError(msg)

    if not result.node.package_name == project.replace("-", "_"):
        msg = f"Wrong project name. Got {result.node.package_name}, expected {project.replace('-', '_')}."
        LOGGER.error(msg)
        raise RuntimeError(msg)

    if not result.node.name == output:
        msg = f"Wrong output name. Got {result.node.name}, expected {output}."
        LOGGER.error(msg)
        raise RuntimeError(msg)

    return result


def get_path(result: RunResult) -> str:
    """
    Get path from dbt result (sql://database/schema/table).

    Parameters
    ----------
    result : RunResult
        The dbt result.

    Returns
    -------
    str
        SQL path.
    """
    try:
        components = result.node.relation_name.replace('"', "")
        components = "/".join(components.split("."))
        return f"sql://{components}"
    except Exception:
        msg = "Something got wrong during path parsing."
        LOGGER.exception(msg)
        raise RuntimeError(msg)


def get_raw_code(result: RunResult) -> str:
    """
    Get raw code from dbt result.

    Parameters
    ----------
    result : RunResult
        The dbt result.

    Returns
    -------
    str
        The raw code.
    """
    try:
        return encode_string(str(result.node.raw_code))
    except Exception:
        msg = "Something got wrong during raw code parsing."
        LOGGER.exception(msg)
        raise RuntimeError(msg)


def get_compiled_code(result: RunResult) -> str:
    """
    Get compiled code from dbt result.

    Parameters
    ----------
    result : RunResult
        The dbt result.

    Returns
    -------
    str
        The compiled code.
    """
    try:
        return encode_string(str(result.node.compiled_code))
    except Exception:
        msg = "Something got wrong during compiled code parsing."
        LOGGER.exception(msg)
        raise RuntimeError(msg)


def get_timings(result: RunResult) -> dict:
    """
    Get timings from dbt result.

    Parameters
    ----------
    result : RunResult
        The dbt result.

    Returns
    -------
    dict
        A dictionary containing timings.
    """
    try:
        compile_timing = None
        execute_timing = None
        for entry in result.timing:
            if entry.name == "compile":
                compile_timing = entry
            elif entry.name == "execute":
                execute_timing = entry
        return {
            "compile": {
                "started_at": compile_timing.started_at.isoformat(),
                "completed_at": compile_timing.completed_at.isoformat(),
            },
            "execute": {
                "started_at": execute_timing.started_at.isoformat(),
                "completed_at": execute_timing.completed_at.isoformat(),
            },
        }
    except Exception:
        msg = "Something got wrong during timings parsing."
        LOGGER.exception(msg)
        raise RuntimeError(msg)



"""
Runtime DBT module.
"""
from __future__ import annotations

import os
import typing
from pathlib import Path

import psycopg2
from dbt.cli.main import dbtRunner, dbtRunnerResult
from digitalhub_core.entities._base.status import State
from digitalhub_core.entities.dataitems.crud import get_dataitem, new_dataitem
from digitalhub_core.runtimes.base import Runtime
from digitalhub_core.utils.exceptions import BackendError, EntityError
from digitalhub_core.utils.generic_utils import build_uuid, decode_string
from digitalhub_core.utils.logger import LOGGER
from digitalhub_core_dbt.runtime.dbt_utils import (
    generate_dbt_profile_yml,
    generate_dbt_project_yml,
    generate_inputs_conf,
    generate_outputs_conf,
)
from digitalhub_core_dbt.runtime.parse_utils import ParsedResults, parse_results
from psycopg2 import sql

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.dataitems.entity import Dataitem


class RuntimeDBT(Runtime):
    """
    Runtime DBT class.
    """

    def __init__(self) -> None:
        """
        Constructor.
        """
        self.root_dir = Path("dbt_run")
        self.model_dir = self.root_dir / "models"
        self._input_dataitems: list[dict[str, str]] = []
        self._table_to_drop: list[str] = []

    def build(self, function: dict, task: dict, run: dict) -> dict:
        """
        Build run spec.

        Parameters
        ----------
        function : dict
            The function.
        task : dict
            The task.
        run : dict
            The run.

        Returns
        -------
        dict
            The run spec.
        """
        return {
            **function.get("spec", {}),
            **task.get("spec", {}),
            **run.get("spec", {}),
        }

    def run(self, run: dict) -> dict:
        """
        Run function.

        Returns
        -------
        dict
            Status of the executed run.
        """

        # Get action
        action = self._get_action(run)

        # Handle unknown task kind
        if action not in ["transform"]:
            msg = f"Task {action} not allowed for DBT runtime"
            LOGGER.error(msg)
            raise EntityError(msg)

        # Execute action
        return self.execute(run)

    ####################
    # TRANSFORM TASK
    ####################

    def execute(self, run: dict) -> dict:
        """
        Execute task.

        Returns
        -------
        dict
            Status of the executed run.
        """

        # Get run specs
        LOGGER.info("Starting task.")
        spec = run.get("spec")
        project = run.get("metadata").get("project")

        # Parse inputs/outputs and decode sql code
        LOGGER.info("Parsing inputs and output.")
        inputs = self._get_inputs(spec.get("inputs", {}).get("dataitems", []), project)
        output = self._get_output_table_name(spec.get("outputs", {}).get("dataitems", []))
        query = self._get_sql(spec)

        # Setup environment
        LOGGER.info("Setting up environment for dbt execution.")
        uuid = build_uuid()
        self.setup(inputs, output, uuid, project, query)

        # Execute function
        LOGGER.info("Executing dbt project.")
        execution_results = self.transform(output)

        # Parse results
        LOGGER.info("Parsing results.")
        parsed_result = parse_results(execution_results, output, project)

        # Create dataitem
        LOGGER.info("Creating output dataitem.")
        dataitem = self._create_dataitem(parsed_result, project, uuid, output)

        # Clean environment
        self.cleanup()

        # Return run status
        LOGGER.info("Task completed, returning run status.")
        return {
            "dataitems": dataitem,
            "timing": parsed_result.timings,
            "state": State.COMPLETED.value,
        }

    ####################
    # Parse inputs/outputs
    ####################

    def _get_inputs(self, inputs: list, project: str) -> list:
        """
        Parse inputs from run spec and materialize dataitems in postgres.

        Parameters
        ----------
        inputs : list
            The list of inputs.
        project : str
            The project name.

        Returns
        -------
        list
            The list of inputs dataitems names.
        """
        for name in inputs:
            di = self._get_dataitem(name, project)
            self._input_dataitems.append({"name": di.metadata.name, "id": di.metadata.version})
            table = self._materialize_dataitem(di, name)
            self._table_to_drop.append(table)
        return inputs

    @staticmethod
    def _get_dataitem(name: str, project: str) -> Dataitem:
        """
        Get dataitem from core.

        Parameters
        ----------
        name : str
            The dataitem name.
        project : str
            The project name.

        Returns
        -------
        Dataitem
            The dataitem.

        Raises
        ------
        BackendError
            If dataitem is not found.
        """
        try:
            LOGGER.info(f"Getting dataitem '{name}'")
            return get_dataitem(project, name)
        except BackendError:
            msg = f"Dataitem {name} not found."
            LOGGER.exception(msg)
            raise BackendError(msg)

    @staticmethod
    def _materialize_dataitem(dataitem: Dataitem, name: str) -> str:
        """
        Materialize dataitem in postgres.

        Parameters
        ----------
        dataitem : Dataitem
            The dataitem.
        name : str
            The dataitem name.

        Returns
        -------
        str
            The materialized table name.

        Raises
        ------
        EntityError
            If something got wrong during dataitem materialization.
        """
        try:
            table_name = f"{name}_v{dataitem.id}"
            LOGGER.info(f"Materializing dataitem '{name}' as '{table_name}'.")
            target_path = f"sql://{os.getenv('POSTGRES_DATABASE')}/{os.getenv('POSTGRES_SCHEMA')}/{table_name}"
            dataitem.write_df(target_path, if_exists="replace")
            return table_name
        except Exception:
            msg = f"Something got wrong during dataitem {name} materialization."
            LOGGER.exception(msg)
            raise EntityError(msg)

    @staticmethod
    def _get_output_table_name(outputs: list) -> str:
        """
        Get output table name from run spec.

        Parameters
        ----------
        outputs : list
            The list of outputs.

        Returns
        -------
        str
            The output dataitem/table name.

        Raises
        ------
        RuntimeError
            If outputs are not a list of one dataitem.
        """
        if not isinstance(outputs, list) or len(outputs) > 1:
            msg = "Outputs must be a list of one dataitem."
            LOGGER.error(msg)
            raise RuntimeError(msg)
        return str(outputs[0])

    def _get_sql(self, spec: dict) -> str:
        """
        Get sql code from run spec.

        Parameters
        ----------
        spec : dict
            The run spec.

        Returns
        -------
        str
            The sql code.

        Raises
        ------
        RuntimeError
            If sql code is not a valid string.
        """
        try:
            return decode_string(spec.get("sql"))
        except Exception:
            msg = "Sql code must be a valid string."
            LOGGER.exception(msg)
            raise RuntimeError(msg)

    ####################
    # Setup environment
    ####################

    def setup(self, inputs: list, output: str, uuid: str, project: str, query: str) -> None:
        """
        Initialize a dbt project with a model and a schema definition.

        Parameters
        ----------
        inputs : list
            The list of inputs.
        output : str
            The output table name.
        uuid : str
            The uuid of the model for outputs versioning.
        project : str
            The project name.
        sql : str
            The sql code.

        Returns
        -------
        None
        """
        # Create directories
        self.model_dir.mkdir(exist_ok=True, parents=True)

        # Generate profile yaml file
        generate_dbt_profile_yml(self.root_dir)

        # Generate project yaml file
        generate_dbt_project_yml(self.root_dir, self.model_dir, project.replace("-", "_"))

        # Generate outputs confs
        generate_outputs_conf(self.model_dir, query, output, uuid)

        # Generate inputs confs for every dataitem
        for di in self._input_dataitems:
            generate_inputs_conf(self.model_dir, di["name"], di["id"])

    ####################
    # Execute function
    ####################

    def transform(self, output: str) -> dbtRunnerResult:
        """
        Execute a dbt project with the specified outputs.
        It initializes a dbt runner, cleans the project and runs it.

        Parameters
        ----------
        output : str
            The output table name.

        Returns
        -------
        dbtRunnerResult
            An object representing the result of the dbt execution.
        """
        os.chdir(self.root_dir)
        dbt = dbtRunner()
        dbt.invoke("clean")
        cli_args = ["run", "--select", f"{output}"]
        res = dbt.invoke(cli_args)
        os.chdir("..")
        return res

    ####################
    # Produce outputs
    ####################

    @staticmethod
    def _create_dataitem(result: ParsedResults, project: str, uuid: str, output: str) -> list[dict]:
        """
        Create new dataitem.

        Parameters
        ----------
        result : ParsedResults
            The parsed results.
        project : str
            The project name.
        uuid : str
            The uuid of the model for outputs versioning.
        output : str
            The output table name.

        Returns
        -------
        list[dict]
            The output dataitem infos.

        Raises
        ------
        RuntimeError
            If something got wrong during dataitem creation.
        """
        try:
            di = new_dataitem(
                project=project,
                name=result.name,
                kind="dataitem",
                path=result.path,
                uuid=uuid,
                raw_code=result.raw_code,
                compiled_code=result.compiled_code,
            )
            return [
                {
                    "key": output,
                    "kind": "dataitem",
                    "id": f"store://{project}/dataitems/dataitem/{di.metadata.name}:{di.metadata.version}",
                }
            ]
        except Exception:
            msg = "Something got wrong during dataitem creation."
            LOGGER.exception(msg)
            raise RuntimeError(msg)

    ####################
    # Cleanup
    ####################

    def cleanup(self) -> None:
        """
        Cleanup environment.

        Returns
        -------
        None
        """
        LOGGER.info("Cleaning up environment.")
        connection = self._get_connection()
        try:
            for table in self._table_to_drop:
                with connection.cursor() as cursor:
                    LOGGER.info(f"Dropping table '{table}'.")
                    query = sql.SQL("DROP TABLE {table}").format(table=sql.Identifier(table))
                    cursor.execute(query)
        except Exception:
            msg = "Something got wrong during environment cleanup."
            LOGGER.exception(msg)
            raise RuntimeError(msg)
        finally:
            LOGGER.info("Closing connection to postgres.")
            connection.close()

    @staticmethod
    def _get_connection() -> psycopg2.extensions.connection:
        """
        Create a connection to postgres and return a session with autocommit enabled.

        Returns
        -------
        psycopg2.extensions.connection
            The connection to postgres.

        Raises
        ------
        RuntimeError
            If something got wrong during connection to postgres.
        """
        try:
            LOGGER.info("Connecting to postgres.")
            connection = psycopg2.connect(
                host=os.getenv("POSTGRES_HOST"),
                port=os.getenv("POSTGRES_PORT"),
                database=os.getenv("POSTGRES_DATABASE"),
                user=os.getenv("POSTGRES_USER"),
                password=os.getenv("POSTGRES_PASSWORD"),
            )
            connection.set_session(autocommit=True)
            return connection
        except Exception:
            msg = "Something got wrong during connection to postgres."
            LOGGER.exception(msg)
            raise RuntimeError(msg)



"""
Task Transform specification module.
"""
from digitalhub_core.entities.tasks.spec import TaskParams, TaskSpec


class TaskSpecTransform(TaskSpec):
    """Task Transform specification."""

    def __init__(
        self,
        function: str,
        **kwargs,
    ) -> None:
        """
        Constructor.
        """
        super().__init__(function, **kwargs)


class TaskParamsTransform(TaskParams):
    """
    TaskParamsTransform model.
    """



from digitalhub_core.utils.modules_utils import ModuleRegistry

registry = ModuleRegistry()
registry.register_runtime("digitalhub_core_dbt.runtime.runtime", "RuntimeDBT")
registry.register_function("digitalhub_core_dbt.entities.functions.spec", "FunctionSpecDBT", "FunctionParamsDBT")
registry.register_tasks(
    "transform", "digitalhub_core_dbt.entities.tasks.spec", "TaskSpecTransform", "TaskParamsTransform"
)
